#include <stdio.h>
#include <stdlib.h>
#include <string.h>

FILE *image;
FILE *output;

unsigned char getbyte() {
	int byte = getc(image);
	if (byte < 0) {
		printf("EOF reached unexpectedly\n");
		exit(1);
	}
	return byte;
}

unsigned short gethalfword() {
	unsigned char hi = getbyte();
	return hi << 8 | getbyte();
}

unsigned int getword() {
	unsigned short hi = gethalfword();
	return hi << 16 | gethalfword();
}

/* Decompression for boot2 */

int getbits(unsigned int n) {
	static unsigned int buf = 0, bits = 0;
	while (bits < n) {
		buf = buf << 8 | getbyte();
		bits += 8;
	}
	bits -= n;
	return buf >> bits & ((1 << n) - 1);
}

void decompress(unsigned int comp_size) {
	unsigned int size = getword();
	unsigned short common[64];
	int i;

	for (i = 0; i < 64; i++)
		common[i] = gethalfword(image);

	for (; size > 0; size -= 2) {
		unsigned short hw;
		if (getbits(1))
			hw = getbits(16);
		else
			hw = common[getbits(6)];
		putc(hw >> 8, output);
		putc(hw,      output);
	}
}

int main(int argc, char **argv) {
	unsigned int flags = 0;

	if (argc < 2) {
		printf("TI-Nspire image decompressor\n"
		       "Usage: IMGDUMP input_file [output_file]\n");

		return 1;
	}

	image = fopen(argv[1], "rb");
	if (!image) {
		perror(argv[1]);
		return 1;
	}

	while (1) {
		unsigned short field = gethalfword();
		unsigned int size = field & 0x000F;
		field &= 0xFFF0;
		if (size == 0x0D)      size = getbyte();
		else if (size == 0x0E) size = gethalfword();
		else if (size == 0x0F) size = getword();

		if (field == 0x8000) {
			/* Don't skip - 8070 is inside this */
		} else if (field == 0x8080) {
			unsigned int addr;
			if (size != 8) {
				printf("Field 8080 has wrong size (%u)\n", size);
				return 1;
			}
			addr = getword();
			flags = getword();
			printf("Base address = 0x%08x, flags = %d\n", addr, flags);
		} else if (field == 0x8070) {
			char *output_filename;
			if (argc > 2) {
				output_filename = argv[2];
			} else {
				output_filename = malloc(strlen(argv[1]) + 5);
				strcpy(output_filename, argv[1]);
				strcat(output_filename, (flags & 4) ? ".zip" : ".raw");
			}
			output = fopen(output_filename, "wb");
			if (!output) {
				perror(output_filename);
				return 1;
			}

			if (flags & 1) {
				decompress(size);
			} else {
				for (; size; size--)
					fputc(getbyte(), output);
			}
		} else if (field == 0xFFF0) {
			break;
		} else {
			/* Skip it */
			fseek(image, size, SEEK_CUR);
		}
	}

	return 0;
}
